/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: automa_management.c,v 1.3 2003/02/20 16:18:07 frassi Exp $
 */


#include <assert.h>
#include "automa_management.h"
#include "util.h"


/******************************************************
 * management functions for the list of automa states *
 ******************************************************/

/*
  init_automa_state

  input:  a list of states of the automa
          a table of variables

  output: a list of states of the automa

  it creates a new element of the list and appends it
*/
automa_state init_automa_state(automa_state csl,var_table_t tab_var)
{
  automa_state_el* new_ase=((automa_state_el*) malloc(sizeof(automa_state_el)));

  csl=NULL;

  new_ase->node_number=0;
  new_ase->tab_var=tab_var;

  csl=g_list_append(csl,new_ase);

  return csl;
}


/*
  free_state_el

  input:  an element of the list of states of the automa
  
  output: nothing

  free all memory allocated for an element of the list
*/
void free_state_el(automa_state_el* s)
{
  free_var_table(s->tab_var);
  free(s);
}


/*
  update_tab_var

  input:  a list of assignments
          a table of variables
	  a trace control block struct

  output: a table of variables

  given the input table of variables and a list of assignments to do, it returns
  an updated table of variables
*/
var_table_t update_tab_var(GList* asg_list,var_table_t tabv,struct tcb* tcb1)
{
  GList *temp_list_el;
  GNode *temp_node;

  temp_list_el=g_list_first(asg_list);
  
  while(temp_list_el!=NULL)
    {
      temp_node=((GNode*) temp_list_el->data);
      ((arg_node_t*) temp_node->data)->funct(temp_node,tabv,tcb1);
      temp_list_el=g_list_next(temp_list_el);
    }
  return tabv;
}

/*
  search state

  input:  a state of the automa (x)
          a list of states of the automa (list)
           
  output: a pointer to the state which is equal to x if it is found in list
          NULL otherwise
 */
GList* search_state(automa_state_el* s,automa_state s_list)
{
  int equal_var_table=0;
  GList* temp_list_el;
  
  temp_list_el=g_list_first(s_list);
  
  while (temp_list_el!=NULL)
    {
      if (s->node_number==((automa_state_el*) temp_list_el->data)->node_number)
	equal_var_table=cmp_var_table_values(s->tab_var,((automa_state_el*) temp_list_el->data)->tab_var);

      if (equal_var_table)
	break;

      temp_list_el=g_list_next(temp_list_el);
    }

  if (equal_var_table)
    return temp_list_el;

  return NULL;
}


/*
  update_state

  input:  an automa
	  a trace control block struct
	  a list of states of the automa

  output: a list of states of the automa

  given the input structures it generates a new list of automa states in which the automa is
  after intercepting a system call
*/
automa_state update_state(automa atm,struct tcb* tcb1,automa_state current_states_list)
{
  int result_cond_test,transition_to_do;
  GList *cur_stat=NULL;
  GList *next_stat=NULL;
  GList *temp=NULL;
  GList *first_trans=NULL;
  GList *save_tab_var=NULL;
  automa_node* atm_node_ptr=NULL;
  edge_t* edge_ptr=NULL;
  struct condition_tree_node* current_cond=NULL;
  //struct condition_tree_node* last_cond=NULL;
  edge_key_t trans_key;

#ifndef NO_CONFLICT
  int i,num_trans;
  automa_state_el *new_cs=NULL;
#endif
  
  /* computes first element of the list (the list can be void) */
  if (current_states_list!=NULL)
    cur_stat=g_list_first(current_states_list);
  else
    cur_stat=NULL;

  /* foreach state S belonging to the list of states... */
  while (cur_stat!=NULL)
    {
      transition_to_do=0;
      edge_ptr=NULL;
      /////////////////////////////////////////tcb1->ev_set=NULL;

      next_stat=g_list_next(cur_stat);

      /* delete S from list of states, preserve the table of variables for future use */
      current_states_list=g_list_remove_link(current_states_list,cur_stat);

      /* I have to find a pointer to the state X in which the automa is in */
      atm_node_ptr=&g_array_index(atm,automa_node,((automa_state_el*) cur_stat->data)->node_number);

      /* prepare the key to look for between the transitions starting from the state of the automa */
      /* This key contains the number of the syscall and the context ("intro" if entering in the system
	 call and "exiting" if, guess what, exiting from the system call) */
      trans_key.syscall_num=tcb1->scno;
      if (entering(tcb1))
	trans_key.context=intro;
      else
	trans_key.context=exiting;

      /*if it exists in state X of the automa a transition caused by the intercepted system call....*/
      edge_ptr=((edge_t*) g_hash_table_lookup(atm_node_ptr->edg_set,&(trans_key)));
      if (edge_ptr!=NULL)
	{
	  /* we have to check if there are conditions on the system call
	     (i.e. system call open with file name == "Mickey Mouse") */
	  current_cond=edge_ptr->ct_root;

	  /*foreach condition C*/
	  while (current_cond->cond_ptr!=NULL)
	    {
	      temp=current_cond->cond_ptr;
	      
	      result_cond_test=1;
	      
	      /* I have to test if the condition is verified */
	      
	      while(temp!=NULL && result_cond_test)
		{
		  /* this is a list of conditions, the result of the test is the && of all results */
		  result_cond_test = result_cond_test && 
		    ((int) ((arg_node_t*) ((GNode*) temp->data)->data)->funct (((GNode*) temp->data),((automa_state_el*) cur_stat->data)->tab_var,tcb1) );
		  
		  temp=g_list_next(temp);
		}
	      
	      if (current_cond->negated_cond)
		result_cond_test=!result_cond_test;
	      
	      if (result_cond_test)
		current_cond=current_cond->cond_pos;
	      else
		current_cond=current_cond->cond_neg;
	      
	      if (!current_cond)
		goto escape;
	    }

	  //last_cond=current_cond;    serve solo a settare ev_set del tcb...

	  /* we found a a transition marked with the system call */
	  transition_to_do=1;
	}
      else 
	{
	  /* 
	     in this case there is no transition marked with the system call we intercepted, starting from
	     state X, therefore we check the existence of a transition marked with DELTA (==any system call)
	  */

	  trans_key.syscall_num=DELTA;
	  trans_key.context=intro;
	  edge_ptr=g_hash_table_lookup(atm_node_ptr->edg_set,&(trans_key));
	  if (edge_ptr==NULL)
	    {
	      trans_key.context=exiting;
	      edge_ptr=g_hash_table_lookup(atm_node_ptr->edg_set,&(trans_key));
	    }

	  if (edge_ptr!=NULL)
	    {
	      current_cond=edge_ptr->ct_root;

	      /****************************/
	      /* we have to check if there are conditions on the system call
		 (i.e. system call open with file name == "Mickey Mouse") */

	      /*foreach condition C*/
	      while (current_cond->cond_ptr!=NULL)
		{
		  temp=current_cond->cond_ptr;
		  
		  result_cond_test=1;
		  
		  /* I have to test if the condition is verified */
		  
		  while(temp!=NULL && result_cond_test)
		    {
		      /* this is a list of conditions, the result of the test is the && of all results */
		      result_cond_test = result_cond_test && 
			((int) ((arg_node_t*) ((GNode*) temp->data)->data)->funct (((GNode*) temp->data),((automa_state_el*) cur_stat->data)->tab_var,tcb1) );
		      temp=g_list_next(temp);
		    }

		  /* negated cond not possible */

		  if (result_cond_test)
		    current_cond=current_cond->cond_pos;
		  else
		    current_cond=current_cond->cond_neg;
		  
		  if (!current_cond)
		    goto escape;
		}
	      /******************************/

	      /* we found a a transition marked with DELTA */
	      transition_to_do=1;
	    }
	  else
	    {
	      /*
		in this case there are no transitions starting froma state X, marked neither with the 
		system call we intercepted, nor with DELTA ===> we can erase S
	      */
	    escape:
	      free_state_el(cur_stat->data);
	      g_list_free_1(cur_stat);
	    }
	}

      if (transition_to_do)
	{
	  first_trans=g_list_first(current_cond->trans_l);

	  /* I update state S ===> now I call it state S1 */
	  ((automa_state_el*) cur_stat->data)->node_number=((single_transition*) first_trans->data)->automa_node_number;
	  
	  /* I update S's table of variables, I also execute any actions caused by the transition 
	     (e.g. terminate the server, log something, etc.) */
	  ((automa_state_el*) cur_stat->data)->tab_var = 
	    update_tab_var(((single_transition*) first_trans->data)->asgs_list,((automa_state_el*) cur_stat->data)->tab_var,tcb1);

	  //////////////////////if (last_cond)
	    /////////////////////tcb1->ev_set=last_cond->ev_set;
	  
	  /*if S1 is not present in the list of states, I append it */
	  if ((g_list_length(current_states_list)==0) || (!(save_tab_var=search_state(((automa_state_el*) cur_stat->data),current_states_list)))) 
	    current_states_list=g_list_concat(cur_stat,current_states_list);
	  else
	    {
	      free_state_el(cur_stat->data);
	      g_list_free_1(cur_stat);
	      cur_stat=save_tab_var;
	    }

/* la definisco se sono sicuro che compilero' file delle specifiche senza conflitti tra assegnamenti */
#ifndef NO_CONFLICT

	  /* there may have been a conflict during the creation of the automa, I treat it now */
	  num_trans=g_list_length(current_cond->trans_l);
	  for (i=1;i<num_trans;i++)
	    {
	      /* I allocate a new element of the list of automa states */
	      new_cs=((automa_state_el*) malloc(sizeof(automa_state_el)));
	      new_cs->tab_var=((var_table_t) clone_var_table(((automa_state_el*) cur_stat->data)->tab_var));
	      temp=g_list_nth(current_cond->trans_l,i);
	      
	      /* I update this element */
	      new_cs->node_number=((single_transition*) temp->data)->automa_node_number;
	      new_cs->tab_var=update_tab_var(((single_transition*) temp->data)->asgs_list,new_cs->tab_var,tcb1);
	      /* I prepend it to the list, otherwise I would update it again on the the next iteration  */
	      if ((g_list_length(current_states_list)==0) || (!(search_state(new_cs,current_states_list)))) 
		current_states_list=g_list_prepend(current_states_list,new_cs);
	      else
		free_state_el(new_cs);
	    }
#endif

	} // end if (transition_to_do)
      
      /* I now evaluate the next element of the list */
      cur_stat=next_stat;
    
    } // end while (cur_stat!=NULL)
  
  return current_states_list;
}

/*
  clone_automa_state_el
  
  input:  an element of the list of automa states
  output: an element of the list of automa states

  given an element of the list, this function clones it and returns a pointer
  to an equal element
*/
automa_state_el *
clone_automa_state_el(automa_state_el *ase)
{
  automa_state_el *new = (automa_state_el *) malloc(sizeof(automa_state_el));
  
  if ( !new )
    {
      fprintf(stderr, "clone_automa_state_el: no memory");
      return NULL;
    }
  
  new->node_number = ase->node_number;
  new->tab_var = clone_var_table(ase->tab_var);
  return new;
}

/*
  clone_automa_state
  
  input:  a list of automa states
  output: a pointer to a list of automa states

  given a list, this function clones it and returns a pointer to an equal list
*/
automa_state *
clone_automa_state(automa_state as)
{
  automa_state_el *new_el;
  automa_state *new_ptr = (automa_state *) malloc(sizeof(automa_state)); 
  
  if (!new_ptr)
    {
      fprintf(stderr, "clone_automa_state: no memory");
      return NULL;
    }
  
  *new_ptr = NULL;
  while ( as )
    {
      new_el = clone_automa_state_el((automa_state_el *) as->data);
      if (!new_el)
	return NULL;
      *new_ptr = g_list_append(*new_ptr, new_el);
      as = g_list_next(as);
    }
  return new_ptr;
}

void
free_atm_state(automa_state process_state)
{
  automa_state tmp;
  
  for (tmp = process_state; tmp; tmp = g_list_next(tmp)) 
    {
      free_var_table(((automa_state_el *)tmp->data)->tab_var);
      free((automa_state_el *)tmp->data);
    }
  
  g_list_free(process_state);
}


/****************************************************
 * DEBUG FUNCTIONS: print the list of automa states *
 ****************************************************/

void 
print_current_state(gpointer acs,gpointer ud)
{
  tprintf(" %d",((automa_state_el*) acs)->node_number);
  print_var_table(((automa_state_el*) acs)->tab_var);
}


void 
print_cs_list(automa_state csl)
{
  tprintf("[automa node: ");
  if (g_list_length(csl) > 0)
    g_list_foreach(csl,print_current_state,NULL);
  else
    tprintf("empty");
  tprintf("]\n");
}



