/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: event_set.c,v 1.1.1.1 2001/12/17 08:58:26 frassi Exp $
 */


#include "event_set.h"

#include <stdlib.h>
#include <limits.h>
#include <assert.h>

/*
 * USO: crea un nuovo event_set allocandolo in memoria
 * INPUT: num      - massimo numero di eventi 
 *        old_set  - NULL oppure un ptr a un insieme di eventi da includere nel nuovo insieme; 
 *                   se il massimo numero di eventi che puo' contenere old_set e' maggiore di num 
 *                   la funzione fallisce e restituisce NULL;  
 * OUTPUT: restituisce un ptr al nuovo event_set allocato 
 */
event_set_t *new_event_set(int num, event_set_t *old_set)
{
  int mod;
  event_set_t *ret = (event_set_t *) malloc(sizeof(event_set_t));
  
  if (!ret)
    return NULL;
 
  ret->num_events = 0;
  ret->iterator = 0;
  ret->max_events = num;
  
  mod = num % (CHAR_BIT * sizeof(char));
  num = num / (CHAR_BIT * sizeof(char)); 
  num = mod ? num + 1 : num; /* num of chars to allocate */

  ret->set = (char *) calloc(num, sizeof(char));

  if (old_set)
    {
      if (num < old_set->max_events)
	{
	  free(ret->set);
	  free(ret);
	  return NULL;
	}
      ret = join_event_set(ret, old_set);
      free(old_set->set);
      free(old_set);
    }
  
  return ret;
}


/*
 * USO: controlla se l'evento indicato e' contenuto nell'event_set  
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato
 *        num_event   - evento da controllare (num_event > 0)
 * OUTPUT: restituisce num_event se la verifica ha esito positivo, 0 altrimenti
 */
int lookup_event(event_set_t *event_set, int num_event)
{
  int num, mod;

  assert(event_set && num_event > 0);

  if (num_event > event_set->max_events)
    return 0;
  num = num_event - 1;

  mod = num % (CHAR_BIT * sizeof(char));
  num = num / (CHAR_BIT * sizeof(char)); 
  
  if (event_set->set[num] & (1 << mod))
    return num_event;

  return 0;
}


/*
 * USO: dealloca l'event set passatogli 
 */
void free_event_set(event_set_t *event_set)
{
  free(event_set->set);
  free(event_set);
}


/*
 * USO: inserisce nell'event_set un evento 
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato
 *        num_event   - evento da aggiungere all'insieme (num_event > 0)
 * OUTPUT: restituisce l'event_set modificato oppure NULL se fallisce;
 */
event_set_t *add_event(event_set_t *event_set, int num_event)
{
  int num, mod;

  assert(event_set && num_event > 0);

  if (num_event > event_set->max_events)
    return NULL;
  num = num_event - 1;

  mod = num % (CHAR_BIT * sizeof(char));
  num = num / (CHAR_BIT * sizeof(char)); 
    
  if (event_set->set[num] & (1 << mod))
    return event_set;
  
  event_set->set[num] |= (1 << mod);
  event_set->num_events++;

  return event_set;
}


/*
 * USO: cancella dall'event_set un evento 
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato
 *        num_event   - evento da togliere all'insieme
 * OUTPUT: restituisce l'event_set modificato oppure NULL se fallisce;
 */
event_set_t *del_event(event_set_t *event_set, int num_event)
{
  int num, mod;

  assert(event_set && num_event > 0);
  
  if (num_event > event_set->max_events)
    return NULL;
  num = num_event - 1;

  mod = num % (CHAR_BIT * sizeof(char));
  num = num / (CHAR_BIT * sizeof(char)); 

  if (event_set->set[num] & (1 << mod))
    { 
      event_set->set[num] &= (1 << mod) ^ UCHAR_MAX;
      event_set->num_events--;
    }
  
  return event_set;
}



/* 
 * USO: esegue l'unione fra due event_set
 * INPUT: a, b   - sono due ptr a event_set precedentemente allocati
 * OUTPUT: restituisce l'event_set dei due in input che puo' contenere il maggior 
 *         numero di eventi nel quale sono stati inseriti gli eventi dell'altro
 * ATTENZIONE: l'event set non modificato non e' deallocato, se lo si volesse fare
 *             occorre prima determinare quale dei due e': se "a" e "b" possono contenere
 *             lo stesso numero di eventi l'event set modificato e' "a" e quello _non_ 
 *             modificato e' "b"; altrimenti e' possibile determinarlo confrontando i puntatori.
 */
event_set_t *join_event_set(event_set_t *a, event_set_t *b)
{
  int i, j, num, mod, min, max, c2t;
  event_set_t *ret; 
  unsigned char and_char;

  assert(a && b);
    
  if (a->max_events >= b->max_events) 
    {
      min = b->max_events;
      max = a->max_events;
      ret = a;
    }
  else
    {
      min = a->max_events;
      max = b->max_events;
      ret = b;
    }
  
  mod = min % (CHAR_BIT * sizeof(char)); /* last offset */
  num = min / (CHAR_BIT * sizeof(char)); 
  num = mod ? num + 1 : num;             /* num of chars to join */

  c2t = 0; /* counted two times */
  for ( i = 0; i < num; i++)
    {
      and_char = a->set[i] & b->set[i];
      if ( and_char )
	{ 
	  /* these subsets aren't disjoint */
	  for ( j = 0; j < CHAR_BIT; j++ )
	    if (and_char & (1 << j))
	      c2t++;
	}
      
      ret->set[i] = a->set[i] | b->set[i];
    }

  ret->max_events = max;
  ret->num_events = a->num_events + b->num_events - c2t;
  
  return ret;
}


/*
 * USO: a partire dalla posizione event_set->iterator restituisce il primo evento che trova
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato;
 *                      event_set->iterator e' di default a zero
 * OUTPUT: restituisce il numero N del primo evento trovato nell'insieme 
 *         (in tal caso event_set->iterator e' posto a N), 
 *         oppure -1 se event_set e' vuoto (event_set->iterator e' resettato a zero) 
 */
int each_event(event_set_t *event_set)
{
  int mod, num;

  assert(event_set && event_set->iterator >= 0);

  for ( ; event_set->iterator < event_set->max_events ; event_set->iterator++ )
    {
      mod = event_set->iterator % (CHAR_BIT * sizeof(char));
      num = event_set->iterator / (CHAR_BIT * sizeof(char)); 
      
      if (event_set->set[num] & (1 << mod))
	return ++event_set->iterator;
    }
  
  event_set->iterator = 0;
  return -1;
}


/*
 * USO: confronta due event set, anche di dimensione diversa
 * INPUT:  a, b   - sono due ptr a event_set precedentemente allocati
 * OUTPUT: restituisce 1 se i due event set hanno la stessa cardinalita' e contengono
 *         gli stessi elementi, 0 altrimenti;
 */
int equal_event_set(event_set_t *a, event_set_t *b)
{
  int mod, num, min, i;
  event_set_t *greatest;

  assert(a && b);

  if (a->num_events != b->num_events)
    return 0;

  if (a->max_events <= b->max_events)
    {
      min = a->max_events;
      greatest = b;
    }
  else
    {
      min = b->max_events;
      greatest = a;
    }
  
  mod = min % (CHAR_BIT * sizeof(char));
  num = min / (CHAR_BIT * sizeof(char)); 
  num = mod ? num + 1 : num; /* num of chars to compare */
  
  for ( i = 0 ; i < num ; i++ )
    if (a->set[i] != b->set[i])
      return 0;
  
  if ( min != greatest->max_events )
    {
      mod = greatest->max_events % (CHAR_BIT * sizeof(char));
      num = greatest->max_events / (CHAR_BIT * sizeof(char)); 
      num = mod ? num + 1 : num; /* num of chars to compare */
      
      for ( ; i < num ; i++ )
	if ( greatest->set[i] != '\0' )
	  return 0;
    }
  
  return 1;
}


int hash_event_set(event_set_t *event_set)
{
  int mod, num, ret;

  assert(event_set);

  mod = event_set->max_events % (CHAR_BIT * sizeof(char));
  num = event_set->max_events / (CHAR_BIT * sizeof(char)); 
  num = mod ? num + 1 : num; /* num of chars to hash */

  ret = 0;
  while ( --num >= 0 )
    ret += event_set->set[num];
  
  return ret;
}

/* end */

