#ifndef EVENT_SET_H
#define EVENT_SET_H

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: event_set.h,v 1.1.1.1 2001/12/17 08:58:48 frassi Exp $
 */


/*************
 * EVENT_SET *
 *************/

typedef struct {
  unsigned char *set;  /* array usato bit a bit */
  int max_events;      /* numero massimo di eventi che puo' contenere l'array */
  int num_events;      /* numero di eventi attualmente contenuti (cardinalita') */
  int iterator;        /* iteratore */
} event_set_t;


#define events_in_set(event_set_ptr)       ((event_set_ptr)->num_events)
/* 
 * USO: restituisce il numero di eventi contenuti in un event_set precedentemente allocato
 */ 


event_set_t *new_event_set(int num, event_set_t *old_set);
/*
 * USO: crea un nuovo event_set allocandolo in memoria
 * INPUT: num      - massimo numero di eventi (num > 0)
 *        old_set  - NULL oppure un ptr a un insieme di eventi da includere nel nuovo insieme; 
                     se l'operazione ha successo old_set viene deallocato;
 *                   se il massimo numero di eventi che puo' contenere old_set e' maggiore di num 
 *                   la funzione fallisce e restituisce NULL;   
 * OUTPUT: restituisce un ptr al nuovo event_set allocato 
 */


void free_event_set(event_set_t *event_set);
/*
 * USO: dealloca l'event set passatogli 
 */


int lookup_event(event_set_t *event_set, int num_event);
/*
 * USO: controlla se l'evento indicato e' contenuto nell'event_set  
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato
 *        num_event   - evento da controllare (num_event > 0)
 * OUTPUT: restituisce num_event se la verifica ha esito positivo, 0 altrimenti
 */


event_set_t *add_event(event_set_t *event_set, int num_event);
/*
 * USO: inserisce nell'event_set un evento e ne incrementa la cardinalita'
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato
 *        num_event   - evento da aggiungere all'insieme (num_event > 0)
 * OUTPUT: restituisce l'event_set modificato oppure NULL se fallisce;
 */


event_set_t *del_event(event_set_t *event_set, int num_event);
/*
 * USO: cancella dall'event_set un evento e ne decrementa la cardinalita'
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato
 *        num_event   - evento da togliere all'insieme (num_event > 0)
 * OUTPUT: restituisce l'event_set modificato oppure NULL se fallisce;
 */


event_set_t *join_event_set(event_set_t *a, event_set_t *b);
/* 
 * USO: esegue l'unione fra due event_set
 * INPUT: a, b   - sono due ptr a event_set precedentemente allocati
 * OUTPUT: restituisce l'event_set dei due in input che puo' contenere il maggior 
 *         numero di eventi nel quale sono stati inseriti gli eventi dell'altro
 * ATTENZIONE: l'event set non modificato non e' deallocato, se lo si volesse fare
 *             occorre prima determinare quale dei due e': se "a" e "b" possono contenere
 *             lo stesso numero di eventi l'event set modificato e' "a" e quello _non_ 
 *             modificato e' "b"; altrimenti e' possibile determinarlo confrontando i puntatori.
 */


#define set_event_set_iterator(event_set, num_event)             ((event_set)->iterator = ((num_event) - 1))
/* 
 * USO: imposta event_set->iterator in modo che la prossima chiamata ad each_event()
 *      restituisca num_event se esiste nell' event set.
 * INPUT:  event_set   - un ptr ad un event_set precedentemente allocato;
 *         num_event   - il primo evento da controllare nella prossima each_event()
 * OUTPUT: restituisce il valore di event_set->iterator
 */
 

int each_event(event_set_t *event_set);
/*
 * USO: a partire dalla posizione event_set->iterator restituisce il primo evento che trova
 *      ( che sara' sicuramente >= (event_set->iterator + 1)).
 * INPUT: event_set   - un ptr ad un event_set precedentemente allocato;
 *                      event_set->iterator e' di default a zero
 * OUTPUT: restituisce il numero N del primo evento trovato nell'insieme 
 *         (in tal caso event_set->iterator e' posto a N), 
 *         oppure -1 se event_set e' vuoto (event_set->iterator e' resettato a zero) 
 *
 * NOTA: questa funzione permette di scorrere uno per uno tutti gli eventi contenuti nell'event_set,
 *       oppure tutti gli eventi a partire da (event_set->iterator + 1)
 */


int equal_event_set(event_set_t *a, event_set_t *b);
/*
 * USO: confronta due event set, anche di dimensione diversa
 * INPUT:  a, b   - sono due ptr a event_set precedentemente allocati
 * OUTPUT: restituisce 1 se i due event set hanno la stessa cardinalita' e contengono
 *         gli stessi elementi, 0 altrimenti;
 */


int hash_event_set(event_set_t *event_set);
/*
 * USO: calcola un valore hash associato all' event_set passatogli come argomento
 * INPUT:  event_set   - un ptr ad un event_set precedentemente allocato;
 * OUTPUT: restituisce un valore intero
 */


#define first_event(event_set_ptr, result) event_set_ptr->iterator=0; result = each_event(event_set_ptr)

#define first_event_after(event_set_ptr,number,result) \
       event_set_ptr->iterator=number; result = each_event(event_set_ptr)



#endif /* ! EVENT_SET_H */



