/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: parser.h,v 1.2 2003/02/17 15:53:53 frassi Exp $
 */

/******************************
 widsard parser: strutture dati
 ******************************/    

#ifndef PARSER_H
#define PARSER_H 

#include <glib.h>
#include <gmodule.h>


/*****************************************
 * File di cui viene eseguito il parsing *
 *****************************************/

char *parsed_file;



/*************************
 * albero della sintassi *
 *************************/

typedef GNode *syntax_tree;     /* GNode tree with event_node_t */ 

#define clone_parse_tree(parse_tree)       clone_tree_with_event_node(parse_tree)
/* clona un albero della sintassi con tutto il suo contenuto */

#define print_parse_tree(parse_tree)       print_tree_with_event_node(parse_tree)
/* stampa un albero della sintassi e tutto il suo contenuto */

#define free_parse_tree(parse_tree)        free_tree_with_event_node(parse_tree)              
/* dealloca un albero della sintassi e tutto il suo contenuto */



/************
 * contesto *
 ************/

typedef enum { intro, exiting } context_t;
/* contesto:
   intro: chiamata alla syscall;
   exiting: ritorno dalla syscall
*/




/**************************************************************************************
 ********************************* TABELLE DEI SIMBOLI ********************************
 **************************************************************************************/


/*************************** 
 * TABELLA DELLE VARIABILI * 
 ***************************/


typedef GHashTable * var_table_t;
typedef GHashTable * macro_var_table_t;

macro_var_table_t macro_var_table;
var_table_t var_table;


/* elemento della tabella */

typedef enum {scalar, hash} var_t;

struct var_entry {
  char *name;        /* nome variabile */
  var_t type_var;    /* hash, scalar */
  int type_values;   /* tipo del valore */
  void *values;      /* un singolo valore o una tabella hash */
};

typedef struct var_entry var_entry;

var_entry *clone_var_entry(var_entry *a);
/* clona la struttura var_entry */

void free_var_entry(var_entry *ve);
/* dealloca la struttura var_entry */

/* * */

#define NEW_VAR_TABLE            g_hash_table_new(g_str_hash, g_str_equal)
/* crea una nuova tabella delle variabili 
   restituisce un oggetto di tipo var_table_t */

#define NEW_MACRO_VAR_TABLE      g_hash_table_new(macro_var_hash, macro_var_equal)
/* crea una nuova tabella delle variabili locali alle macro 
   restituisce un oggetto di tipo macro_var_table_t */

#define putvar(var_table, var_name, var_entry_ptr)      g_hash_table_insert(var_table, var_name, var_entry_ptr)
/* inserisce un elemento nella tabella delle variabili:
   var_name e' il nome della variabile
   var_attr e' un puntatore ad una struttura var_entry */

#define getvar(var_table, var_name)                     ((var_entry *) g_hash_table_lookup(var_table, var_name))
/* restituisce se esiste un puntatore alla struttura var_entry associata alla variabile var_name 
   nella tabella delle variabili, NULL altrimenti */

var_table_t clone_var_table(var_table_t var_table);
/* clona la tabella delle variabili ma non la tabella delle variabili delle macro */

void free_var_table(var_table_t var_table);
/* dealloca la tabella delle variabili */

void putmacrovar(macro_var_table_t var_table, char *var_name, int macro_id, var_entry *var_attr);
/* inserisce un elemento nella tabella delle variabili locali alle macro:
   var_name e' il nome della variabile, 
   macro_id e' l'identificatore della macro in cui la variabile e' definita */

var_entry *getmacrovar(macro_var_table_t var_table, char *var_name, int macro_id);
/* restituisce se esiste un puntatore alla struttura var_entry associata alla chiave (var_name, macro_id)
   nella tabella delle variabili locali alle macro, NULL altrimenti */

GList *getmacrovarlist(macro_var_table_t var_table, int macro_id);
/* restituisce un lista di strutture var_entry ciascuna delle quali ha come chiave (* , macro_id) */

void print_var_table(var_table_t var_table);
/* stampa i nomi delle variabili contenute in una tabella delle variabili oppure in una tabella 
   di variabili locali alle macro */

void free_macro_var_table(macro_var_table_t macro_var_table);
/* dealloca la tabella di variabili locali alle macro */

int cmp_var_table_values(var_table_t first, var_table_t second);
/* controlla che le tabelle contengano gli stessi elementi e ne confronta i valori;
   restituisce 1 se le tabelle sono uguali, 0 altrimenti */

/* * */
guint macro_var_hash(gconstpointer key);
gboolean macro_var_equal(gconstpointer a, gconstpointer b);
/* * */




/********************************************
 * Tabella delle visiblita' delle variabili *
 ********************************************

 La tabella delle visibilita' e' una tabella hash di coppie (key, val).
 key contiene il nome della variabile e il numero della sequenza (o regola) in cui 
 la variabile e' definita. val contiene una lista (ordinata) di elementi scope_el_t.
 
 La tabella utilizzata tenendo conto della seguente convenzione:              
 - il numero di sequenza (num_seq) e' negativo per le definizioni di macro                   
 - num_seq e' uguale a zero per le variabili definite globalmente 
 - num_seq e' positivo per le sequenze normali

*/



/* elemento della tabella */

struct scope_el_t {   
  int num_event;       /* inizializzato a -1 */
  context_t context;  
  int num_statement; 
};

typedef struct scope_el_t scope_el_t;

/* * */


/* funzioni per gestire la tabella */

void init_var_scope_table();
/* Inizializza la tabella delle visibilita' delle variabili */

scope_el_t *putscope(char *name, int num_seq);
/*
  Restituisce il puntatore ad una struttura scope_el_t associata alla coppia (name, num_seq).
  Alla stessa coppia (name, num_seq) possono essere associati pi elementi di tipo scope_el_t.
  ATTENZIONE: Se si vuole mantenere la lista delle strutture scope_el_t ordinata,
  dopo aver impostato i campi della struttura restituita utilizzare la funzione
  sort_current_scope()
*/

scope_el_t *getscope(char *name, int num_seq, unsigned int n);
/* 
  Restituisce il puntatore alla "n"-esima struttura scope_el_t associata alla coppia di 
  valori (name, num_seq) purche' esista, altrimenti restituisce NULL. 
  ATTENZIONE: n parte da 1.
*/

#define firstscope(name, num_seq)   getscope(name, num_seq, 1)
/* 
  Restituisce il puntatore alla prima struttura scope_el_t associata alla coppia di 
  valori (name, num_seq) purche' esista, altrimenti restituisce NULL.
*/

scope_el_t *lastscope(char *name, int num_seq);
/* 
  Restituisce il puntatore all'ultima struttura scope_el_t associata alla coppia di 
  valori (name, num_seq) purche' esista, altrimenti restituisce NULL.
*/

scope_el_t *sort_current_scope(void);
/*
  Sposta l'elemento contenente l'ultima struttura scope_el_t restituita da una delle
  precedenti funzioni, in maniera che la lista delle strutture scope_el_t sia mantenuta
  ordinata. Restituisce l'elemento corrente.
*/

scope_el_t *nextscope(void);
/*
  Restituisce il puntatore alla struttura scope_el_t successiva all'ultima
  struttura restituita da una delle seguenti funzioni:
  putscope(), getscope(), firstscope(), lastscope(), nextscope().
*/

scope_el_t *prevscope(void);
/*
  Restituisce il puntatore alla struttura scope_el_t precedente all'ultima
  struttura restituita da una delle seguenti funzioni:
  putscope(), getscope(), firstscope(), lastscope(), nextscope().
*/






/**************************
 * TABELLE DEI SEGNAPOSTI *
 **************************/

typedef GHashTable * label_table_t;


label_table_t macro_label_table;
label_table_t label_table;


/* coppia di elementi della tabella */

typedef struct {
  char *name;
  int num_event;
} label_key_t;

typedef struct {
  label_key_t *key;
  int type_label;
  int n_arg;             /* compreso tra 0 e (arg_number - 1)
			    se == rval il segnaposto e' messo come valore di ritorno della syscall
			    se == -1 il segnaposto e' contenuto in una condizione o in una azione
			    associata ad una combinazione di eventi */
} label_value_t;

/* * */

#define NEW_LABEL_TABLE             g_hash_table_new(label_hash, label_equal)
/* crea una nuova tabella di segnaposti
   restituisce un oggetto di tipo label_table_t */

label_value_t *put_label(label_table_t table, char *label_name, int num_event);  
/* inserisce un nuovo elemento della tabella con chiave (label_name, num_event);
   restituisce un puntatore alla struttura label_value_t associata alla chiave
   (label_name, num_event). La struttura label_value_t e' inizializzata con valori 
   di default ed in particolare con il valore -1 per il campo n_arg */

label_value_t *get_label(label_table_t table, char *label_name, int num_event); 
/* restituisce, se esiste, un puntatore alla struttura label_value_t associata 
   alla chiave (label_name, num_event), NULL altrimenti */

void free_label_table(label_table_t table);
/* delloca la tabella passatagli */

/* * */
guint label_hash(gconstpointer key);
gboolean label_equal(gconstpointer a, gconstpointer b);
/* * */



/***********************
 * TABELLA DELLE MACRO *
 ***********************/


/* elemento della tabella */

struct macro_entry {
  char *name;
  int macro_id;
  syntax_tree macro_val;

  struct macro_entry *next;
};

typedef struct macro_entry macro_entry;

/* * */

macro_entry *macro_table;

/* * */

macro_entry *putmacro(char *macro_name, int macro_id, macro_entry **table);
/* mette una macro nella lista puntata da table della relativa tabella */

macro_entry *getmacro_by_name(char *macro_name, macro_entry **table); 
/* restituisce la macro dalla lista puntata da table della relativa tabella */

/* TODO 
macro_entry *getmacro_by_id(int macro_id, macro_entry **table);
* come la precedente ma la ricerca avviene per identificatore
*/

void print_macro_entries(macro_entry *table);
/* stampa la tabella delle macro */

void free_macro_table(macro_entry *table);
/* dealloca la tabella delle macro */




/**************************
 * TABELLA DELLE FUNZIONI *
 **************************/


typedef GHashTable * funct_table_t;

funct_table_t funct_table;

/* elemento tabella */

struct funct_entry {
  char *name;
  char *so_filename;   /* shared object file name */
  int args_num;
  int ret_type;
  int args_type[10]; 
  //void *ptr;
  void * (* ptr)();
};

typedef struct funct_entry funct_entry;


void init_funct_table(funct_table_t *funct_table);
/* crea una nuova tabella delle funzioni inizializzandola con quelle predefinite */
   
#define putfunct(funct_table, funct_name, funct_entry_ptr)      g_hash_table_insert(funct_table, funct_name, funct_entry_ptr)
/* inserisce un elemento nella tabella delle funzioni:
   funct_name e' il nome della funzione
   funct_entry_ptr e' un puntatore ad una struttura funct_entry allocata */

#define getfunct(funct_table, funct_name)                       ((funct_entry *) g_hash_table_lookup(funct_table, funct_name))
/* restituisce se esiste un puntatore alla struttura funct_entry associata alla variabile funct_name 
   nella tabella delle variabili, NULL altrimenti */

int funct_ret_type ( void *(*funct_ptr)() );
/* dato un puntatore a funzione che si trova in funct_predefined[] restituisce
   il tipo di ritorno della medesima funzione */


/**********************
 * TABELLA DEI MODULI *
 **********************/

typedef GHashTable * module_table_t;

module_table_t module_table;

struct module_entry {
  char *so_filename;
  GModule *libhandle;
};

typedef struct module_entry module_entry;

#define NEW_MODULE_TABLE            g_hash_table_new(g_str_hash, g_str_equal)
/* crea una nuova tabella dei moduli da caricare a runtime e 
   restituisce un oggetto di tipo module_table_t */

#define putmodule(module_table, module_name, module_entry_ptr)      g_hash_table_insert(module_table, module_name, module_entry_ptr)
/* inserisce un elemento nella tabella dei moduli da caricare a runtime:
   module_name e' il nome della libreria condivisa (shared object)
   module_entry_ptr e' un puntatore ad una struttura module_entry allocata */

#define getmodule(module_table, module_name)                       ((module_entry *) g_hash_table_lookup(module_table, module_name))
/* restituisce se esiste un puntatore alla struttura funct_entry associata alla variabile funct_name 
   nella tabella delle variabili, NULL altrimenti */





/************************************************************************************************
 *************************** ELEMENTI SINTATTICI DEL PARSE TREE *********************************
 ************************************************************************************************/


/***********
 * HANDLER *
 ***********/


typedef struct {
  char *var_name;        /* nome della variabile, se != NULL */
  long type_label;        

  union {              
    long n_arg;          /* numero arg dell'evento n_sysc, se var_name == NULL */ 
    char *label_name;    /* nome del segnaposto, se var_name == NULL */
  } label_ref; 

} handler_t;


handler_t *clone_handler(handler_t *handler);
/* clona l'handler */ 

void print_handler(handler_t *a);
/* stampa il contenuto della struttura handler_t */




/*************
 * CONTAINER *
 *************/


typedef enum {HANDLER, SIMPLE_VALUE, HASH_VALUE, ARRAY} content_t;


typedef struct {

  content_t type_content;
  int type_value; 

  union { 
    handler_t *handler;
    GHashTable *hash_value;
    void *value;  
  } content;

} container;




container *clone_container(container *a);
/* clona il container e le strutture ad esso collegate */

void print_container(container *a);
/* stampa il contenuto del container e delle strutture ad esso collegate */

void free_container(container *a);
/* dealloca il container e le strutture ad esso collegate */




/* funzioni per il contenuto del container */

 char *clone_string(char *a);
/* clona una stringa */

char *clone_char(char *a);
/* clona un carattere */

long *clone_long(long *a);
/* clona un long int */

GList *clone_array(GList *a);
/* clona un array */

void print_array(GList *a);
/* stampa il contenuto di un array */

void free_array(GList *a);
/* dealloca un array */

GHashTable *clone_hash_table(GHashTable *a, int type_value);
/* clona una tabella hash contenente valori di tipo type_value */

void print_hash_table(GHashTable *a, int type_value);
/* stampa il contenuto di una hashtable contenente valori di tipo type_value */

void free_hash_table(GHashTable *a);
/* dealloca la tabella hash e tutti i valori in essa contenuti */

int cmp_hash_tables(GHashTable *a, GHashTable *b);
/* confronta il contenuto delle due hash table;
   resituisce 1 se sono uguali, 0 altrimenti */

/* * */



/************
 * arg_node *
 ************/


typedef struct {
  container *arg;         /* if arg == null then funct != NULL */ 
  void *(* funct)();      /* if funct == null then arg != NULL 
			     (durante il parsing c'e' il puntatore alla funct_entry associata alla funzione 
			     mentre a runtime ho il puntatore vero e proprio alla funzione) */
} arg_node_t;


arg_node_t *clone_arg_node(arg_node_t *a);
/* clona la struttura di tipo arg_node_t e le strutture ad essa collegate */

GNode *clone_tree_with_arg_node(GNode *a);
/* clona un albero o un ramo contenente strutture di tipo arg_node_t;
   la funzione richiama la funzione clone_arg_node */

void free_tree_with_arg_node(GNode *a);
/* dealloca un albero o un ramo contenente strutture di tipo arg_node_t;
   la funzione dealloca gli arg_node e le strutture ad essi collegate */

void print_tree_with_arg_node(GNode *a);
/* stampa il contenuto di un albero di arg_node */




/****************
 * simple_event *
 ****************/

/* valori speciali per n_sysc */
#define ANY        -1     /* indica una qualsiasi chiamata di sistema */
#define DELTA      -2     /* indica una qualsiasi chiamata di sistema fra quelle che non compaiono nel file delle regole */
#define END_MARKER -3     /* indica la fine di una regola */


typedef struct {

  int n_sysc;             /* numero della chiamata di sistema; puo' assumere i valori speciali indicati sopra */
  int num_event;          /* identificatore univoco dell'evento */
  context_t context;      
  
  int empty_follow;       /* se empty_follow == 1 allora la funzione follow applicata al simple_event corrente 
			     restituisce un insieme vuoto */

  union { 
    int negated_condition;     /* se negated_condition == 1 allora le condizioni dell'evento sono negate */
    GList *minus_sysc;         /* se n_sysc == ANY questo campo puo' avere la lista delle syscall da non includere */
  } fl;
  
  GList *condition;       /* GNode list with arg_node_t */
  GList *action;          /* GNode list with arg_node_t */
} simple_event_t;



simple_event_t *clone_simple_event(simple_event_t *a);
/* clona il simple_event e tutte le strutture ad esso collegate */

void print_simple_event(simple_event_t *a);
/* stampa il simple_event e tutte le strutture ad esso collegate */

void free_simple_event(simple_event_t *a);
/* dealloca un simple_event */



/* funzioni per il contenuto di un simple_event */

GList *clone_condition_or_action(GList *a);
/* clona una lista di condizioni o di azioni (statements) */

void print_condition_or_action(GList *a);
/* stampa una lista di condizioni o di azioni (statements) */

void free_condition_or_action(GList *a);
/* dealloca una lista di condizioni o di azioni (statements) */

/* * */




/**************
 * event_node *
 **************/


typedef enum { none, non_occurence, sequencing, alternation, repetition } event_operator;

typedef struct {
  event_operator op;
  simple_event_t *simple_event;
} event_node_t;


event_node_t *clone_event_node(event_node_t *a);
/* clona un event node e le strutture ad esso collegate */

void print_event_node(event_node_t *a);
/* stampa un event node e le strutture ad esso collegate */

GNode *clone_tree_with_event_node(GNode *a);
/* clona un albero con event node e tutto cio' che contiene */

void print_tree_with_event_node(GNode *a);
/* stampa un albero con event node e tutto cio' che contiene */

void free_tree_with_event_node(GNode *a);      
/* dealloca un albero con event node e tutto cio' che contiene */







/********************************************************************************************************
 ************************************************* VARIE ************************************************
 ********************************************************************************************************/



/****************************************************************************
 * funzioni che restituiscono info sulle syscall (definite in pars_funct.c) *
 ****************************************************************************


int find_sysc_number(char *sysc_name); 
 * cerca la chiamata di sistema nella tabella definita in syscallent.h *

int get_nargs(int num_sysc);
 * dato il num. di syscall restituisce il num. di argomenti che ha *

int is_intro(int sysc_num, int args_number);
 * dato il num. di syscall e il num. di argomento restituisce se e' INTRO o EXIT *

char *get_sysc_name(int num_sysc);
* dato il num. di una syscall restituisce il corrispondente nome */




/********************************************************************
 * funzioni definite per la gestione dei flag (definite in flags.c) *
 ********************************************************************/

int lookup_flag(char *name); 					    



#endif /* !PARSER_H */
